package com.quart {
	import caurina.transitions.Tweener;

	import flash.events.KeyboardEvent;

	import com.bit101.components.HUISlider;
	import com.bit101.components.Window;
	import com.uncontrol.BoilerPlate;

	import org.papervision3d.Papervision3D;
	import org.papervision3d.cameras.Camera3D;
	import org.papervision3d.core.math.Number3D;
	import org.papervision3d.materials.ColorMaterial;
	import org.papervision3d.materials.WireframeMaterial;
	import org.papervision3d.objects.DisplayObject3D;
	import org.papervision3d.objects.primitives.Plane;
	import org.papervision3d.render.BasicRenderEngine;
	import org.papervision3d.scenes.Scene3D;
	import org.papervision3d.view.Viewport3D;

	import flash.display.Sprite;
	import flash.display.StageQuality;
	import flash.display.StageScaleMode;
	import flash.events.Event;
	import flash.events.MouseEvent;

	/**
	 * @author a-mata
	 */
	[SWF(width="400", height="400", backgroundColor="#EEEEEE", frameRate="60")]

	public class Spinner extends BoilerPlate {
		private var STAGE_WIDTH : Number;
		private var STAGE_HEIGHT : Number;
		private var viewport : Viewport3D;
		private var base : DisplayObject3D;
		private var cameraTarget : DisplayObject3D;
		private var renderer : BasicRenderEngine;
		private var scene : Scene3D;
		private var camera : Camera3D;

		private var isOrbiting : Boolean = false;
		private var previousMouseX : Number;
		private var previousMouseY : Number;

		// private var isKeyPressed : Boolean = false;
		private var TWO_PI : Number = Math.PI * 2;

		private var cameraPitch : Number;
		private var cameraYaw : Number;
		private var plane : Plane;

		private var totalIncA : int = 3;
		private var totalIncB : int = 40;
		private var planeDataArray : Array;
		private var transformed3DArray : Array;
		private var planeArray : Array;
		private var ribbonOffsetSpeed : Number = 0;
		private var sliderHUISliderArray : Array;
		private var ribbonWidthSlider : HUISlider;
		private var ribbonSpeedSlider : HUISlider;
		private var ribbonOffsetIncSlider : HUISlider;
		private var swirlOffsetSlider : HUISlider;
		private var swirlSpanSlider : HUISlider;
		private var swirlSizeSlider : HUISlider;

		public function Spinner() {
			setScope('Spinner');
			this.addEventListener(Event.ADDED_TO_STAGE, init, false, 0, true);
		}

		public function init(e : Event = null) : void {
			tracer('init');
			this.removeEventListener(Event.ADDED_TO_STAGE, init);
			STAGE_WIDTH = stage.stageWidth;
			STAGE_HEIGHT = stage.stageHeight;

			// stage.displayState 	= StageDisplayState.FULL_SCREEN;
			stage.scaleMode = StageScaleMode.NO_SCALE;
			stage.quality = StageQuality.HIGH;
			
			createWorld();
			// createGridMarks();
			createControls();
			createShapes();
			
			onMouseDownHandler();
			onMouseMoveHandler();
			onMouseUpHandler();
			
			stage.addEventListener(KeyboardEvent.KEY_UP, onKeyUpHandler, false, 0, true);
			this.addEventListener(Event.ENTER_FRAME, update, false, 0, true);
		}

		private function onKeyUpHandler(e : KeyboardEvent) : void {
			// tracer('onKeyUpHandler');
			if(e.keyCode == 32) {
				setRandomControlPositions();
			}		
		}

		private function setRandomControlPositions():void {
			
			var time:Number = 5.0;
			
			for (var i:int = 0; i < sliderHUISliderArray.length; i++ ) {
				var slider:HUISlider = sliderHUISliderArray[i];
				Tweener.addTween(slider, { value:(slider.maximum-slider.minimum)*Math.random() + slider.minimum, time:time, transition:'easeInOutQuad' } );
			}
			
		}
		
		/**
		 * create world sets the defaults for all papervision info
		 * base is the displayObject3D that everything is applied to similiar to this.context
		 * baseOffset is used for and world transformations 
		 */
		private function createWorld() : void {
			tracer('createGridMarks');
			
			var square : Sprite = new Sprite();
			square.graphics.beginFill(0xFFFFFF);
			square.graphics.drawRect(0, 0, STAGE_WIDTH, STAGE_HEIGHT);
			square.graphics.endFill();
			square.addEventListener(MouseEvent.MOUSE_DOWN, onMouseDownHandler, false, 0, true);
			square.addEventListener(MouseEvent.MOUSE_MOVE, onMouseMoveHandler, false, 0, true);
			square.addEventListener(MouseEvent.MOUSE_UP, onMouseUpHandler, false, 0, true);
			square.addEventListener(MouseEvent.MOUSE_WHEEL, onMouseWheelHandler, false, 0, true);
			
			this.addChild(square);
			
			Papervision3D.useDEGREES = true;
			viewport = new Viewport3D(STAGE_WIDTH, STAGE_HEIGHT, false, true, true);
			viewport.containerSprite.name = 'paperBaseView';						// name required for hit with construction
			viewport.interactive = true;
			square.addChild(viewport);
			
			renderer = new BasicRenderEngine();
			scene = new Scene3D();
			base = new DisplayObject3D();
			scene.addChild(base);
			
			cameraPitch = 47;
			cameraYaw = 288;
			camera = new Camera3D();
			
			cameraTarget = new DisplayObject3D();
			base.addChild(cameraTarget);
			
			camera.target = cameraTarget;
		}

		/**
		 * create XYZ marks for easy viewing
		 */
		private function createGridMarks() : void {
			tracer('createGridMarks');
			var grid3D : DisplayObject3D = new DisplayObject3D();
			var clutMat : WireframeMaterial = new WireframeMaterial();
			var cursor : Plane;

			clutMat = new WireframeMaterial(0xFF0000, 1);
			clutMat.doubleSided = true;
			cursor = new Plane(clutMat, 1, 500, 1, 1);
			grid3D.addChild(cursor);
			
			clutMat = new WireframeMaterial(0x00FF00, 1);
			clutMat.doubleSided = true;
			cursor = new Plane(clutMat, 1, 500, 1, 1);
			cursor.rotationX = 90;
			grid3D.addChild(cursor);
			
			clutMat = new WireframeMaterial(0x0000FF, 1);
			clutMat.doubleSided = true;

			cursor = new Plane(clutMat, 1, 500, 1, 1);
			cursor.rotationX = 90;
			cursor.rotationY = 90;

			grid3D.addChild(cursor);
			base.addChild(grid3D);
		}

		/**
		 * create slider controls
		 */
		private function createControls() : void {
			sliderHUISliderArray = new Array();
			var yOffset : Number = 10;
			var leftMargin : Number = 8;
			var sliderWidth : Number = 340;
			
			var sliderWindow : Window = new Window(this, 10, 10, 'ADVANCED\t\t\t\t\t\t\t Use spacebar to randomize values');
			sliderWindow.width = 380;
			sliderWindow.height = 140;
			sliderWindow.color = 0xFFFFFF;
			sliderWindow.hasMinimizeButton = true;
			sliderWindow.shadow = false;
			sliderWindow.minimized = true;

			ribbonWidthSlider = new HUISlider(sliderWindow.content, leftMargin, yOffset, "ribbonWidthSlider");
			ribbonWidthSlider.minimum = 0;
			ribbonWidthSlider.maximum = 200;
			ribbonWidthSlider.value = 20;
			ribbonWidthSlider.tick = .001;
			sliderHUISliderArray.push(ribbonWidthSlider);
			yOffset += 20;

			ribbonSpeedSlider = new HUISlider(sliderWindow.content, leftMargin, yOffset, "ribbonSpeedSlider");
			ribbonSpeedSlider.minimum = 0;
			ribbonSpeedSlider.maximum = 50;
			ribbonSpeedSlider.value = 10;
			ribbonSpeedSlider.tick = .001;
			sliderHUISliderArray.push(ribbonSpeedSlider);
			yOffset += 20;

			ribbonOffsetIncSlider = new HUISlider(sliderWindow.content, leftMargin, yOffset, "ribbonOffsetIncSlider");
			ribbonOffsetIncSlider.minimum = 0;
			ribbonOffsetIncSlider.maximum = 2;
			ribbonOffsetIncSlider.value = .5;
			ribbonOffsetIncSlider.tick = .001;
			sliderHUISliderArray.push(ribbonOffsetIncSlider);
			yOffset += 20;
			
			yOffset += 20;
			
			swirlOffsetSlider = new HUISlider(sliderWindow.content, leftMargin, yOffset, "swirlOffsetSlider");
			swirlOffsetSlider.minimum = -720;
			swirlOffsetSlider.maximum = 720;
			swirlOffsetSlider.value = 360;
			swirlOffsetSlider.tick = .001;
			sliderHUISliderArray.push(swirlOffsetSlider);
			yOffset += 20;
			
			swirlSpanSlider = new HUISlider(sliderWindow.content, leftMargin, yOffset, "swirlSpanSlider");
			swirlSpanSlider.minimum = 0;
			swirlSpanSlider.maximum = 360;
			swirlSpanSlider.value = 180;
			swirlSpanSlider.tick = .001;
			sliderHUISliderArray.push(swirlSpanSlider);
			yOffset += 20;
			
			swirlSizeSlider = new HUISlider(sliderWindow.content, leftMargin, yOffset, "swirlSizeSlider");
			swirlSizeSlider.minimum = 0;
			swirlSizeSlider.maximum = 400;
			swirlSizeSlider.value = 200;
			swirlSizeSlider.tick = .001;
			sliderHUISliderArray.push(swirlSizeSlider);
			yOffset += 20;
			
			for (var i : int = 0;i < sliderHUISliderArray.length;i++ ) {
				var slider : HUISlider = sliderHUISliderArray[i];
				// slider.label = '';
				slider.width = sliderWidth;
			}
			
			yOffset += 20;
			sliderWindow.height = yOffset + 20;
		}

		/**
		 * create shapes based on primitives
		 */
		private function createShapes() : void {
			tracer('createShapes');
			
			// create shape
			var wireframeMaterial : WireframeMaterial;
			/*
			var wireframeMaterial : WireframeMaterial = new WireframeMaterial(0x000000, .25);
			wireframeMaterial.doubleSided = true;
			wireframeMaterial.interactive = false;
			 */
			var colorMaterial : ColorMaterial;
			/*
			var colorMaterial : ColorMaterial = new ColorMaterial(0x000000, 1.0);
			colorMaterial.doubleSided = true;
			colorMaterial.interactive = false;
			 */

			planeArray = new Array();
			
			for (var k : int = 0;k < 10;k++) {
				
				colorMaterial = new ColorMaterial(0xFFFFFF * k / 10, 1.0);
				colorMaterial.doubleSided = true;
				colorMaterial.interactive = false;
				
				wireframeMaterial = new WireframeMaterial(0xFFFFFF * k / 10, 1.0);
				wireframeMaterial.doubleSided = true;
				wireframeMaterial.interactive = false;
				
				plane = new Plane(colorMaterial, 50, 500, totalIncA - 1, totalIncB - 1);
				base.addChild(plane);
				planeArray.push(plane);
			}
			
			
			
			// create data for shape
			var i : uint;
			var j : uint;
			var seed : int = 0;
			planeDataArray = new Array();
			
			for (j = 0;j < totalIncA;j++) {
				for (i = 0;i < totalIncB;i++) {
					planeDataArray[seed] = new Number3D(0, 0, 0);
					seed++;
				}
			}
		}

		/**
		 * loop via Event.ENTER_FRAME
		 */
		private function update(e : Event = null) : void {
			var i : uint;
			var j : uint;
			var k : uint;
			var seed : int;
			var geometrySeed : int;

			var oldNum3D : Number3D;
			
			var anchorA : Number3D = new Number3D(0, 0, 0);
			var transformA : Number3D;
			
			var anchorB : Number3D = new Number3D(0, 0, 0);
			var transformB : Number3D;
			
			var anchorC : Number3D = new Number3D(0, 0, 0);
			var transformC : Number3D;

			transformed3DArray = new Array();

			// parse
			seed = 0;
			for (j = 0;j < totalIncA;j++) {
				for (i = 0;i < totalIncB;i++) {
					
					oldNum3D = planeDataArray[seed];
					oldNum3D.x = 0;//i / (totalIncB - 1) * 800;
					oldNum3D.y = -300;
					var ribbonWidth : Number = Math.cos((i + ribbonOffsetSpeed) / ribbonSpeedSlider.value * TWO_PI) * ribbonWidthSlider.value + ribbonWidthSlider.value;
					ribbonWidth = (1 - i / (totalIncB - 1)) * ribbonWidth;
					
					oldNum3D.z = j / (totalIncA - 1) * ribbonWidth - ribbonWidth / 2;
					
					transformed3DArray[seed] = oldNum3D;
					
					/*
					if(i < 5) {
					transformA = getRotationAroundAnchorZ(oldNum3D, anchorA, i / 5 * mouseX);
					} else {
					oldNum3D.x = (i - 5) / (totalIncB - 1) * 400;
					transformA = getRotationAroundAnchorZ(oldNum3D, anchorA, mouseX);
					}
					transformed3DArray[seed] = transformA;
					 */

					transformA = getRotationAroundAnchorZ(oldNum3D, anchorA, i / (totalIncB - 1) * swirlSpanSlider.value);
					transformed3DArray[seed] = transformA;
					
					transformB = getRotationAroundAnchorY(transformA, anchorB, (1 - i / (totalIncB - 1)) * swirlOffsetSlider.value);
					transformed3DArray[seed] = transformB;
					
					
					transformC = getScalarAroundAnchor(transformB, anchorC, swirlSizeSlider.value / 200);
					transformC.x = transformB.x;
					transformC.z = transformB.z;

					transformed3DArray[seed] = transformC;
					
					// swirlSizeSlider.value/400;

					seed++;
				}
			}
			ribbonOffsetSpeed += ribbonOffsetIncSlider.value;
			
			
			// draw
			seed = 0;
			geometrySeed = 0;
			for (j = 0;j < totalIncA;j++) {
				for (i = 0;i < totalIncB;i++) {
					geometrySeed = i + j * totalIncB;
					
					for (k = 0;k < 10;k++) {
						plane = planeArray[k];

						plane.geometry.vertices[geometrySeed].x = transformed3DArray[seed].x;	
						plane.geometry.vertices[geometrySeed].y = transformed3DArray[seed].y;	
						plane.geometry.vertices[geometrySeed].z = transformed3DArray[seed].z;
						
						plane.rotationY = 360 / 10 * k;
						// plane.y = k/10*200;
					}					

					seed++;
				}
			}
			
			base.rotationY += 1;
			renderer.renderScene(scene, camera, viewport);							// standard loop required to render papervision
		}

		
		
		
		
		
		
		// =====================================================================
		// mouse handlers
		// =====================================================================

		private function onMouseWheelHandler(event : MouseEvent = null) : void {
			camera.moveForward(30 * event.delta);
		}

		private function onMouseDownHandler(event : MouseEvent = null) : void {
			isOrbiting = true;
			previousMouseX = mouseX;
			previousMouseY = mouseY;
			onMouseMoveHandler();
		}

		private function onMouseUpHandler(event : MouseEvent = null) : void {
			isOrbiting = false;
		}

		private function onMouseMoveHandler(event : MouseEvent = null) : void {
			var differenceX : Number = mouseX - previousMouseX;
			var differenceY : Number = mouseY - previousMouseY;

			if (isOrbiting) {
				cameraPitch += differenceY;
				cameraYaw += differenceX;
		
				cameraPitch %= 360;
				cameraYaw %= 360;
				
				cameraPitch = cameraPitch < 90 ? cameraPitch : 89.9999;
				cameraPitch = cameraPitch > 0 ? cameraPitch : 0.0001;

				previousMouseX = mouseX;
				previousMouseY = mouseY;

				camera.orbit(cameraPitch, cameraYaw);
			}
		}

		
		
		
		
		
		
		// =====================================================================
		// anchor transforms 
		// =====================================================================

		public function getRotationAroundAnchorZ(oldPoint3D : Number3D, anchorPoint3D : Number3D, degrees : Number = 0) : Number3D {
			var newPoint3D : Number3D = Number3D.sub(oldPoint3D.clone(), anchorPoint3D);
			newPoint3D.rotateZ(degrees);
			return Number3D(Number3D.add(newPoint3D, anchorPoint3D));
		}

		public function getRotationAroundAnchorX(oldPoint3D : Number3D, anchorPoint3D : Number3D, degrees : Number = 0) : Number3D {
			var newPoint3D : Number3D = Number3D.sub(oldPoint3D.clone(), anchorPoint3D);
			newPoint3D.rotateX(degrees);
			return Number3D(Number3D.add(newPoint3D, anchorPoint3D));
		}

		public function getRotationAroundAnchorY(oldPoint3D : Number3D, anchorPoint3D : Number3D, degrees : Number = 0) : Number3D {
			var newPoint3D : Number3D = Number3D.sub(oldPoint3D.clone(), anchorPoint3D);
			newPoint3D.rotateY(degrees);
			return Number3D(Number3D.add(newPoint3D, anchorPoint3D));
		}

		public function getScalarAroundAnchor(oldPoint3D : Number3D, anchorPoint3D : Number3D, ratio : Number = 0) : Number3D {
			var newPoint3D : Number3D = Number3D.sub(oldPoint3D.clone(), anchorPoint3D);
			newPoint3D.multiplyEq(ratio);
			return Number3D(Number3D.add(newPoint3D, anchorPoint3D));
		}

		public function getTranslationAroundAnchor(oldPoint3D : Number3D, anchorPoint3D : Number3D) : Number3D {
			return Number3D(Number3D.add(oldPoint3D.clone(), anchorPoint3D));
		}
	}
}